@@ -3,7 +3,7 @@
  *
  * See the IPFILTER.LICENCE file for details on licencing.
  *
- * $Id: ip_auth.h,v 2.3.2.6 2002/10/26 07:03:00 darrenr Exp $
+ * $Id: ip_auth.h,v 2.16 2003/07/25 12:29:56 darrenr Exp $
  *
  */
 #ifndef	__IP_AUTH_H__
@@ -13,10 +13,12 @@
 
 typedef struct  frauth {
 	int	fra_age;
+	int	fra_len;
 	int	fra_index;
 	u_32_t	fra_pass;
 	fr_info_t	fra_info;
-#if SOLARIS
+	char	*fra_buf;
+#ifdef	MENTAT
 	queue_t	*fra_q;
 #endif
 } frauth_t;
@@ -44,20 +46,19 @@ typedef struct  fr_authstat {
 extern	frentry_t	*ipauth;
 extern	struct fr_authstat	fr_authstats;
 extern	int	fr_defaultauthage;
+extern	int	fr_authstart;
+extern	int	fr_authend;
 extern	int	fr_authsize;
 extern	int	fr_authused;
 extern	int	fr_auth_lock;
-extern	u_32_t	fr_checkauth __P((ip_t *, fr_info_t *));
+extern	frentry_t *fr_checkauth __P((fr_info_t *, u_32_t *));
 extern	void	fr_authexpire __P((void));
+extern	int	fr_authinit __P((void));
 extern	void	fr_authunload __P((void));
-extern	mb_t	*fr_authpkts[];
-extern	int	fr_newauth __P((mb_t *, fr_info_t *, ip_t *));
-#if defined(__NetBSD__) || defined(__OpenBSD__) || \
-    (__FreeBSD_version >= 300003)
-extern	int	fr_preauthcmd __P((u_long, frentry_t *, frentry_t **));
-extern	int	fr_auth_ioctl __P((caddr_t, int, u_long));
-#else
-extern	int	fr_preauthcmd __P((int, frentry_t *, frentry_t **));
-extern	int	fr_auth_ioctl __P((caddr_t, int, int));
-#endif
+extern	int	fr_authflush __P((void));
+extern	mb_t	**fr_authpkts;
+extern	int	fr_newauth __P((mb_t *, fr_info_t *));
+extern	int	fr_preauthcmd __P((ioctlcmd_t, frentry_t *, frentry_t **));
+extern	int	fr_auth_ioctl __P((caddr_t, ioctlcmd_t, int));
+
 #endif	/* __IP_AUTH_H__ */
