@@ -1,5 +1,20 @@
+#include <linux/version.h>
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+# define __irq_h        1       /* stop it being included! */
+#else
+# define _LINUX_TCP_H
+#endif
+#include <net/ip.h>
+
 #include "ipf-linux.h"
 
+#include <net/checksum.h>
+#include <net/route.h>
+
+#include <linux/random.h>
+#include <linux/timer.h>
+#include <asm/ioctls.h>
+
 extern int sysctl_ip_default_ttl;
 
 static	int	frzerostats __P((caddr_t));
@@ -7,11 +22,46 @@ static	int	fr_send_ip __P((fr_info_t *, struct sk_buff *, struct sk_buff **));
 
 ipfmutex_t	ipl_mutex, ipf_authmx, ipf_rw, ipf_stinsert;
 ipfmutex_t	ipf_nat_new, ipf_natio, ipf_timeoutlock;
-ipfrwlock_t	ipf_mutex, ipf_global, ipf_ipidfrag;
+ipfrwlock_t	ipf_mutex, ipf_global, ipf_ipidfrag, ipf_frcache;
 ipfrwlock_t	ipf_frag, ipf_state, ipf_nat, ipf_natfrag, ipf_auth;
+struct timer_list	ipf_timer;
 
 static u_int ipf_linux_inout __P((u_int, struct sk_buff **, const struct net_device *, const struct net_device *, int (*okfn)(struct sk_buff *)));
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+static struct nf_hook_ops ipf_hooks[] = {
+	{
+		.hook		= ipf_linux_inout,
+		.owner		= THIS_MODULE,
+		.pf		= PF_INET,
+		.hooknum	= NF_IP_PRE_ROUTING,
+		.priority	= 200,
+	},
+	{
+		.hook		= ipf_linux_inout,
+		.owner		= THIS_MODULE,
+		.pf		= PF_INET,
+		.hooknum	= NF_IP_POST_ROUTING,
+		.priority	= 200,
+	},
+# ifdef USE_INET6
+	{
+		.hook		= ipf_linux_inout,
+		.owner		= THIS_MODULE,
+		.pf		= PF_INET6,
+		.hooknum	= NF_IP_PRE_ROUTING,
+		.priority	= 200,
+	},
+	{
+		.hook		= ipf_linux_inout,
+		.owner		= THIS_MODULE,
+		.pf		= PF_INET6,
+		.hooknum	= NF_IP_POST_ROUTING,
+		.priority	= 200,
+	}
+# endif
+};
+#else
 static struct	nf_hook_ops	ipf_hooks[] = {
 	{
 		{ NULL, NULL },		/* list */
@@ -27,7 +77,7 @@ static struct	nf_hook_ops	ipf_hooks[] = {
 		NF_IP_POST_ROUTING,	/* hooknum */
 		200			/* priority */
 	},
-#ifdef USE_INET6
+# ifdef USE_INET6
 	{
 		{ NULL, NULL },		/* list */
 		ipf_linux_inout,	/* hook */
@@ -42,8 +92,96 @@ static struct	nf_hook_ops	ipf_hooks[] = {
 		NF_IP_POST_ROUTING,	/* hooknum */
 		200			/* priority */
 	}
-#endif
+# endif
 };
+#endif
+
+
+int iplattach()
+{
+	int err, i;
+
+	SPL_NET(s);
+	if (fr_running > 0) {
+		SPL_X(s);
+		return -EBUSY;
+	}
+
+	bzero((char *)frcache, sizeof(frcache));
+	MUTEX_INIT(&ipf_rw, "ipf rw mutex");
+	MUTEX_INIT(&ipl_mutex, "ipf log mutex");
+	MUTEX_INIT(&ipf_timeoutlock, "ipf timeout lock mutex");
+	RWLOCK_INIT(&ipf_global, "ipf global rwlock");
+	RWLOCK_INIT(&ipf_mutex, "ipf global mutex rwlock");
+	RWLOCK_INIT(&ipf_frcache, "ipf cache mutex rwlock");
+	RWLOCK_INIT(&ipf_ipidfrag, "ipf IP NAT-Frag rwlock");
+
+	for (i = 0; i < sizeof(ipf_hooks)/sizeof(ipf_hooks[0]); i++) {
+		err = nf_register_hook(&ipf_hooks[i]);
+		if (err != 0)
+			return err;
+	}
+
+	if (fr_initialise() < 0) {
+		for (i = 0; i < sizeof(ipf_hooks)/sizeof(ipf_hooks[0]); i++)
+			nf_unregister_hook(&ipf_hooks[i]);
+		SPL_X(s);
+		return EIO;
+	}
+
+	bzero((char *)frcache, sizeof(frcache));
+#ifdef notyet
+	if (fr_control_forwarding & 1)
+		ipv4_devconf.forwarding = 1;
+#endif
+
+	SPL_X(s);
+	/* timeout(fr_slowtimer, NULL, (hz / IPF_HZ_DIVIDE) * IPF_HZ_MULT); */
+	init_timer(&ipf_timer);
+	ipf_timer.function = fr_slowtimer;
+	ipf_timer.data = NULL;
+	ipf_timer.expires = (HZ / IPF_HZ_DIVIDE) * IPF_HZ_MULT;
+	add_timer(&ipf_timer);
+	mod_timer(&ipf_timer, HZ/2 + jiffies);
+	return 0;
+}
+
+
+int ipldetach()
+{
+	int i;
+
+	del_timer(&ipf_timer);
+
+	SPL_NET(s);
+
+	for (i = 0; i < sizeof(ipf_hooks)/sizeof(ipf_hooks[0]); i++)
+		nf_unregister_hook(&ipf_hooks[i]);
+	/* untimeout(fr_slowtimer, NULL); */
+
+#ifdef notyet
+	if (fr_control_forwarding & 2)
+		ipv4_devconf.forwarding = 0;
+#endif
+
+	fr_deinitialise();
+
+	(void) frflush(IPL_LOGIPF, 0, FR_INQUE|FR_OUTQUE|FR_INACTIVE);
+	(void) frflush(IPL_LOGIPF, 0, FR_INQUE|FR_OUTQUE);
+
+	MUTEX_DESTROY(&ipf_timeoutlock);
+	MUTEX_DESTROY(&ipl_mutex);
+	MUTEX_DESTROY(&ipf_rw);
+	RW_DESTROY(&ipf_mutex);
+	RW_DESTROY(&ipf_frcache);
+	RW_DESTROY(&ipf_global);
+	RW_DESTROY(&ipf_ipidfrag);
+
+	SPL_X(s);
+
+	return 0;
+}
+
 
 
 /*
@@ -64,7 +202,8 @@ int ipf_ioctl(struct inode *in, struct file *fp, u_int cmd, u_long arg)
 		if (unit != IPL_LOGIPF)
 			return -EIO;
 		if (cmd != SIOCIPFGETNEXT && cmd != SIOCIPFGET &&
-		    cmd != SIOCIPFSET && cmd != SIOCFRENB && cmd != SIOCGETFS)
+		    cmd != SIOCIPFSET && cmd != SIOCFRENB &&
+		    cmd != SIOCGETFS && cmd != SIOCGETFF)
 			return -EIO;
 	}
 
@@ -172,10 +311,21 @@ int ipf_ioctl(struct inode *in, struct file *fp, u_int cmd, u_long arg)
 			error = EPERM;
 		else {
 			bcopy(data, &tmp, sizeof(tmp));
-			tmp = frflush(unit, tmp);
+			tmp = frflush(unit, 4, tmp);
 			bcopy(&tmp, data, sizeof(tmp));
 		}
 		break;
+#ifdef USE_INET6
+	case	SIOCIPFL6 :
+		if (!(mode & FWRITE))
+			error = EPERM;
+		else {
+			bcopy(data, &tmp, sizeof(tmp));
+			tmp = frflush(unit, 6, tmp);
+			bcopy(&tmp, data, sizeof(tmp));
+		}
+		break;
+#endif
 	case SIOCSTLCK :
 		error = COPYIN(data, &tmp, sizeof(tmp));
 		if (error == 0) {
@@ -201,7 +351,7 @@ int ipf_ioctl(struct inode *in, struct file *fp, u_int cmd, u_long arg)
 		if (!(mode & FWRITE))
 			error = EPERM;
 		else {
-			frsync();
+			frsync(NULL);
 		}
 		break;
 	default :
@@ -245,6 +395,9 @@ fr_info_t *fin;
 	if (fr_checkl4sum(fin) == -1)
 		return -1;
 #endif
+	m = skb_copy(fin->fin_m, GFP_ATOMIC);
+	if (m == NULL)
+		return -1;
 
 	tlen = (tcp->th_flags & (TH_SYN|TH_FIN)) ? 1 : 0;
 #ifdef	USE_INET6
@@ -254,13 +407,13 @@ fr_info_t *fin;
 #endif
 		hlen = sizeof(ip_t);
 	hlen += sizeof(*tcp2);
-	m = alloc_skb(hlen + 16, GFP_ATOMIC);
-	if (m == NULL)
-		return -1;
+	skb_trim(m, hlen);
 
-	m->data += 16;
 	bzero(MTOD(m, char *), hlen);
-	tcp2 = (struct tcphdr *)(MTOD(m, char *) + hlen - sizeof(*tcp2));
+	ip = MTOD(m, ip_t *);
+	bzero((char *)ip, hlen);
+	ip->ip_v = fin->fin_v;
+	tcp2 = (tcphdr_t *)((char *)ip + hlen - sizeof(*tcp2));
 	tcp2->th_dport = tcp->th_sport;
 	tcp2->th_sport = tcp->th_dport;
 	if (tcp->th_flags & TH_ACK) {
@@ -272,15 +425,11 @@ fr_info_t *fin;
 		tcp2->th_ack = htonl(tcp2->th_ack);
 		tcp2->th_flags = TH_RST|TH_ACK;
 	}
-	tcp2->th_off = sizeof(struct tcphdr) >> 2;
+	tcp2->th_off = sizeof(*tcp2) >> 2;
 
-	/*
-	 * This is to get around a bug in the Solaris 2.4/2.5 TCP checksum
-	 * computation that is done by their put routine.
-	 */
 #ifdef	USE_INET6
 	if (fin->fin_v == 6) {
-		ip6 = MTOD(m, ip6_t *);
+		ip6 = (ip6_t *)ip;
 		ip6->ip6_src = fin->fin_dst6;
 		ip6->ip6_dst = fin->fin_src6;
 		ip6->ip6_plen = htons(sizeof(*tcp));
@@ -288,14 +437,11 @@ fr_info_t *fin;
 	} else
 #endif
 	{
-		ip = MTOD(m, ip_t *);
+		ip->ip_hl = sizeof(*ip) >> 2;
 		ip->ip_src.s_addr = fin->fin_daddr;
 		ip->ip_dst.s_addr = fin->fin_saddr;
-		ip->ip_id = fr_nextipid(fin);
-		ip->ip_hl = sizeof(*ip) >> 2;
 		ip->ip_p = IPPROTO_TCP;
-		ip->ip_len = htons(sizeof(*ip) + sizeof(*tcp));
-		ip->ip_tos = fin->fin_ip->ip_tos;
+		ip->ip_len = sizeof(*ip) + sizeof(*tcp);
 		tcp2->th_sum = fr_cksum(m, ip, IPPROTO_TCP, tcp2);
 	}
 	return fr_send_ip(fin, m, &m);
@@ -306,26 +452,187 @@ static int fr_send_ip(fin, sk, skp)
 fr_info_t *fin;
 struct sk_buff *sk, **skp;
 {
+	fr_info_t fnew;
 	ip_t *ip, *oip;
+	int hlen;
 
 	ip = MTOD(sk, ip_t *);
+	bzero((char *)&fnew, sizeof(fnew));
 	oip = fin->fin_ip;
 
-	ip->ip_v = fin->fin_v;
 	switch (fin->fin_v)
 	{
 	case 4 :
+		fnew.fin_v = 4;
 		ip->ip_hl = sizeof(*oip) >> 2;
 		ip->ip_tos = oip->ip_tos;
-		ip->ip_id = oip->ip_id;
-		ip->ip_off = 0;
+		ip->ip_id = 0;
 		ip->ip_ttl = sysctl_ip_default_ttl;
 		ip->ip_sum = 0;
+		ip->ip_off = 0x4000;
+		hlen = sizeof(*ip);
 		break;
 	default :
 		return EINVAL;
 	}
-	return fr_fastroute(sk, skp, fin, NULL);
+
+	fnew.fin_ifp = fin->fin_ifp;
+	fnew.fin_flx = FI_NOCKSUM;
+	fnew.fin_m = sk;
+	fnew.fin_ip = ip;
+	fnew.fin_mp = skp;
+	fnew.fin_hlen = hlen;
+	fnew.fin_dp = (char *)ip + hlen;
+	(void) fr_makefrip(hlen, ip, &fnew);
+
+	return fr_fastroute(sk, skp, &fnew, NULL);
+}
+
+
+int fr_send_icmp_err(type, fin, isdst)
+int type;
+fr_info_t *fin;
+int isdst;
+{
+	int hlen, code, leader, dlen;
+	struct in_addr dst4;
+	struct icmp *icmp;
+	u_short sz;
+#ifdef	USE_INET6
+	ip6_t *ip6;
+	mb_t *mb;
+#endif
+	ip_t *ip;
+	mb_t *m, *m0;
+
+	if ((type < 0) || (type > ICMP_MAXTYPE))
+		return -1;
+
+	code = fin->fin_icode;
+
+#ifdef USE_INET6
+	if ((code < 0) || (code > sizeof(icmptoicmp6unreach)/sizeof(int)))
+		return -1;
+#endif
+
+#ifndef	IPFILTER_CKSUM
+	if (fr_checkl4sum(fin) == -1)
+		return -1;
+#endif
+
+	m0 = fin->fin_m;
+#ifdef USE_INET6
+	if (fin->fin_v == 6) {
+		sz = sizeof(ip6_t);
+		dlen = MIN(M_LEN(m0), 512);
+		sz += dlen;
+		hlen = sizeof(ip6_t);
+		type = icmptoicmp6types[type];
+		if (type == ICMP6_DST_UNREACH)
+			code = icmptoicmp6unreach[code];
+	} else
+#endif
+	{
+		if ((fin->fin_p == IPPROTO_ICMP) && !(fin->fin_flx & FI_SHORT))
+			switch (ntohs(fin->fin_data[0]) >> 8)
+			{
+			case ICMP_ECHO :
+			case ICMP_TSTAMP :
+			case ICMP_IREQ :
+			case ICMP_MASKREQ :
+				break;
+			default :
+				return 0;
+			}
+
+		sz = sizeof(ip_t) * 2 + 4;
+		dlen = 8;		/* 64 bits of data */
+		sz += dlen;
+		hlen = sizeof(ip_t);
+	}
+
+	leader = m0->data - m0->head;
+	if ((leader & 15) != 0)
+		leader += 16 - (leader & 15);
+        m = alloc_skb(sz + leader, GFP_ATOMIC);
+        if (m == NULL)
+                return -1;
+                
+        /* Set the data pointer */
+        skb_reserve(m, leader);
+
+	bzero(MTOD(m, char *), (size_t)sz);
+
+	m->nh.iph = (struct iphdr *)skb_put(m, hlen);
+	ip = (ip_t *)m->nh.iph;
+	ip->ip_v = fin->fin_v;
+
+	m->h.icmph = (struct icmphdr *)skb_put(m, hlen + 4 + dlen);
+	icmp = (icmphdr_t *)m->h.icmph;
+	icmp->icmp_type = type & 0xff;
+	icmp->icmp_code = code & 0xff;
+#ifdef	icmp_nextmtu
+	if (type == ICMP_UNREACH && (qif->qf_max_frag != 0) &&
+	    fin->fin_icode == ICMP_UNREACH_NEEDFRAG)
+		icmp->icmp_nextmtu = htons(qif->qf_max_frag);
+#endif
+
+#ifdef	USE_INET6
+	if (fin->fin_v == 6) {
+		struct in6_addr dst6;
+		int csz;
+
+		if (isdst == 0) {
+			if (fr_ifpaddr(6, FRI_NORMAL, qif->qf_ill,
+				       (struct in_addr *)&dst6, NULL) == -1) {
+				FREE_MB_T(m);
+				return -1;
+			}
+		} else
+			dst6 = fin->fin_dst6;
+
+		csz = sz;
+		sz -= sizeof(ip6_t);
+		ip6 = (ip6_t *)ip;
+		ip6->ip6_plen = htons((u_short)sz);
+		ip6->ip6_nxt = IPPROTO_ICMPV6;
+		ip6->ip6_src = dst6;
+		ip6->ip6_dst = fin->fin_src6;
+		sz -= offsetof(struct icmp, icmp_ip);
+		bcopy((char *)mb->b_rptr, (char *)&icmp->icmp_ip, sz);
+		icmp->icmp_cksum = csz - sizeof(ip6_t);
+	} else
+#endif
+	{
+		ip = MTOD(m, ip_t *);
+		ip->ip_hl = sizeof(*ip) >> 2;
+		ip->ip_p = IPPROTO_ICMP;
+		ip->ip_len = (u_short)sz;
+		if (isdst == 0) {
+			if (fr_ifpaddr(4, FRI_NORMAL, fin->fin_ifp,
+				       &dst4, NULL) == -1) {
+				FREE_MB_T(m);
+				return -1;
+			}
+		} else
+			dst4 = fin->fin_dst;
+		ip->ip_src = dst4;
+		ip->ip_dst = fin->fin_src;
+		bcopy((char *)fin->fin_ip, (char *)&icmp->icmp_ip,
+		      sizeof(*fin->fin_ip));
+		icmp->icmp_ip.ip_len = htons(icmp->icmp_ip.ip_len);
+		icmp->icmp_ip.ip_off = htons(icmp->icmp_ip.ip_off);
+		bcopy((char *)fin->fin_ip + fin->fin_hlen,
+		      (char *)&icmp->icmp_ip + sizeof(*fin->fin_ip), 8);
+		icmp->icmp_cksum = ip_compute_csum((u_char *)icmp,
+						   sz - sizeof(ip_t));
+	}
+
+	/*
+	 * Need to exit out of these so we don't recursively call rw_enter
+	 * from fr_qout.
+	 */
+	return fr_send_ip(fin, m, &m);
 }
 
 
@@ -352,156 +659,154 @@ fr_info_t *fin;
 }
 
 
-int fr_send_icmp_err(type, fin, isdst)
-int type;
-fr_info_t *fin;
-int isdst;
-{
-	struct in_addr dst4;
-	struct icmp *icmp;
-	int hlen, code;
-	u_short sz;
-#ifdef	USE_INET6
-	ip6_t *ip6;
-	mb_t *mb;
-#endif
-	ip_t *ip;
-	mb_t *m;
-
-	if ((type < 0) || (type > ICMP_MAXTYPE))
-		return -1;
-
-	code = fin->fin_icode;
-#ifdef USE_INET6
-	if ((code < 0) || (code > sizeof(icmptoicmp6unreach)/sizeof(int)))
-		return -1;
-#endif
-
-#ifndef	IPFILTER_CKSUM
-	if (fr_checkl4sum(fin) == -1)
-		return -1;
-#endif
-
-#ifdef	USE_INET6
-	mb = fin->fin_qfm;
-
-	if (fin->fin_v == 6) {
-		sz = sizeof(ip6_t);
-		sz += MIN(mb->b_wptr - mb->b_rptr, 512);
-		hlen = sizeof(ip6_t);
-		type = icmptoicmp6types[type];
-		if (type == ICMP6_DST_UNREACH)
-			code = icmptoicmp6unreach[code];
-	} else
-#endif
-	{
-		if ((fin->fin_p == IPPROTO_ICMP) && !(fin->fin_flx & FI_SHORT))
-			switch (ntohs(fin->fin_data[0]) >> 8)
-			{
-			case ICMP_ECHO :
-			case ICMP_TSTAMP :
-			case ICMP_IREQ :
-			case ICMP_MASKREQ :
-				break;
-			default :
-				return 0;
-			}
-
-		sz = sizeof(ip_t) * 2;
-		sz += 8;		/* 64 bits of data */
-		hlen = sizeof(ip_t);
-	}
-
-	sz += offsetof(struct icmp, icmp_ip);
-	m = alloc_skb(hlen + 16, GFP_ATOMIC);
-	if (m == NULL)
-		return -1;
-	m->data += 16;
-	bzero(MTOD(m, char *), (size_t)sz);
-	icmp = (struct icmp *)(MTOD(m, char *) + hlen);
-	icmp->icmp_type = type & 0xff;
-	icmp->icmp_code = code & 0xff;
-#ifdef	icmp_nextmtu
-	if (type == ICMP_UNREACH && (qif->qf_max_frag != 0) &&
-	    fin->fin_icode == ICMP_UNREACH_NEEDFRAG)
-		icmp->icmp_nextmtu = htons(qif->qf_max_frag);
-#endif
-
-#ifdef	USE_INET6
-	if (fin->fin_v == 6) {
-		struct in6_addr dst6;
-		int csz;
-
-		if (isdst == 0) {
-			if (fr_ifpaddr(6, FRI_NORMAL, qif->qf_ill,
-				       (struct in_addr *)&dst6, NULL) == -1)
-				return -1;
-		} else
-			dst6 = fin->fin_dst6;
-
-		csz = sz;
-		sz -= sizeof(ip6_t);
-		ip6 = (ip6_t *)m->b_rptr;
-		ip6->ip6_plen = htons((u_short)sz);
-		ip6->ip6_nxt = IPPROTO_ICMPV6;
-		ip6->ip6_src = dst6;
-		ip6->ip6_dst = fin->fin_src6;
-		sz -= offsetof(struct icmp, icmp_ip);
-		bcopy((char *)mb->b_rptr, (char *)&icmp->icmp_ip, sz);
-		icmp->icmp_cksum = csz - sizeof(ip6_t);
-	} else
-#endif
-	{
-		ip = MTOD(m, ip_t *);
-		ip->ip_hl = sizeof(*ip) >> 2;
-		ip->ip_p = IPPROTO_ICMP;
-		ip->ip_id = fin->fin_ip->ip_id;
-		ip->ip_tos = fin->fin_ip->ip_tos;
-		ip->ip_len = htons((u_short)sz);
-		if (isdst == 0) {
-			if (fr_ifpaddr(4, FRI_NORMAL, fin->fin_ifp,
-				       &dst4, NULL) == -1)
-				return -1;
-		} else
-			dst4 = fin->fin_dst;
-		ip->ip_src = dst4;
-		ip->ip_dst = fin->fin_src;
-		bcopy((char *)fin->fin_ip, (char *)&icmp->icmp_ip,
-		      sizeof(*fin->fin_ip));
-		bcopy((char *)fin->fin_ip + fin->fin_hlen,
-		      (char *)&icmp->icmp_ip + sizeof(*fin->fin_ip), 8);
-		icmp->icmp_cksum = ipf_cksum((u_short *)icmp,
-					     sz - sizeof(ip_t));
-	}
-
-	/*
-	 * Need to exit out of these so we don't recursively call rw_enter
-	 * from fr_qout.
-	 */
-	return fr_send_ip(fin, m, &m);
-}
-
-
 u_short fr_nextipid(fin)
 fr_info_t *fin;
 {
+#if 1
+	static u_short ipid = 0;
+
+	return ipid++;
+#else
 	ip_t ip;
 
 	__ip_select_ident(&ip, NULL);
 	return ip.ip_id;
+#endif
 }
 
 
 /*ARGSUSED*/
-int fr_fastroute(min, mp, fin, dst)
+int fr_fastroute(min, mp, fin, fdp)
 mb_t *min, **mp;
 fr_info_t *fin;
-frdest_t *dst;
+frdest_t *fdp;
 {
-	int result;
+	struct net_device *ifp, *sifp;
+	struct in_addr dip;
+	struct rtable *rt;
+	frentry_t *fr;
+	int err, sout;
+	ip_t *ip;
 
-	result = ip_queue_xmit(min);
-	return result;
+	rt = NULL;
+	fr = fin->fin_fr;
+	ip = MTOD(min, ip_t *);
+	dip = ip->ip_dst;
+
+	if (fdp != NULL)
+		ifp = fdp->fd_ifp;
+	else
+		ifp = fin->fin_ifp;
+
+	if ((ifp == NULL) && ((fr == NULL) || !(fr->fr_flags & FR_FASTROUTE))) {
+		err = ENETUNREACH;
+		goto bad;
+	}
+
+	/*
+	 * In case we're here due to "to <if>" being used with "keep state",
+	 * check that we're going in the correct direction.
+	 */
+	if ((fr != NULL) && (fin->fin_rev != 0)) {
+		if ((ifp != NULL) && (fdp == &fr->fr_tif))
+			return -1;
+	}
+	if (fdp != NULL) {
+		if (fdp->fd_ip.s_addr)
+			dip = fdp->fd_ip;
+	}
+
+	switch (fin->fin_v)
+	{
+	case 4 :
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
+		err = ip_route_output(&rt, dip.s_addr, 0,
+				       RT_TOS(ip->ip_tos) | RTO_CONN, 0);
+#else
+		err = 1;
+#endif
+		if (err != 0 || rt == NULL)
+			goto bad;
+
+		if (rt->u.dst.dev == NULL) {
+			err = EHOSTUNREACH;
+			goto bad;
+		}
+		break;
+	default :
+		err = EINVAL;
+		goto bad;
+	}
+
+	if (fin->fin_out == 0) {
+		sifp = fin->fin_ifp;
+		sout = fin->fin_out;
+		fin->fin_ifp = ifp;
+		fin->fin_out = 1;
+		(void) fr_acctpkt(fin, NULL);
+		fin->fin_fr = NULL;
+		if (!fr || !(fr->fr_flags & FR_RETMASK)) {
+			u_32_t pass;
+
+			(void) fr_checkstate(fin, &pass);
+		}
+
+		switch (fr_checknatout(fin, NULL))
+		{
+		case 0 :
+			break;
+		case 1 :
+			ip->ip_sum = 0;
+			break;
+		case -1 :
+			err = EINVAL;
+			goto bad;
+			break;
+		}
+
+		fin->fin_ifp = sifp;
+		fin->fin_out = sout;
+	}
+	ip->ip_sum = 0;
+
+
+	if (min->dst != NULL) {
+		dst_release(min->dst);
+		min->dst = NULL;
+	}
+
+	min->dst = &rt->u.dst;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,21)
+	if (min->len > min->dst->pmtu) {
+		err = EMSGSIZE;
+		goto bad;
+	}
+#endif
+
+	switch (fin->fin_v)
+	{
+	case 4 :
+		ip->ip_len = htons(ip->ip_len);
+		ip->ip_off = htons(ip->ip_off);
+		ip->ip_sum = ip_fast_csum((u_char *)ip, ip->ip_hl);
+
+		/*dumpskbuff(min);*/
+		NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, min, NULL,
+			ifp, ip_finish_output);
+		err = 0;
+		break;
+
+	default :
+		err = EINVAL;
+		break;
+	}
+
+	if (err == 0)
+		return 0;
+bad:
+	if (min != NULL)
+		kfree_skb(min);
+	return err;
 }
 
 
@@ -515,6 +820,9 @@ struct in_addr *inp, *inpmask;
 	struct net_device *dev;
 	struct in_device *ifp;
 
+	if ((ifptr == NULL) || (ifptr == (void *)-1))
+		return -1;
+
 	dev = ifptr;
 	ifp = __in_dev_get(dev);
 
@@ -575,84 +883,6 @@ caddr_t	data;
 }
 
 
-int iplattach()
-{
-	int err, i;
-
-	SPL_NET(s);
-	if (fr_running > 0) {
-		SPL_X(s);
-		return -EBUSY;
-	}
-
-	bzero((char *)frcache, sizeof(frcache));
-	MUTEX_INIT(&ipf_rw, "ipf rw mutex");
-	MUTEX_INIT(&ipl_mutex, "ipf log mutex");
-	MUTEX_INIT(&ipf_timeoutlock, "ipf timeout lock mutex");
-	RWLOCK_INIT(&ipf_global, "ipf global rwlock");
-	RWLOCK_INIT(&ipf_mutex, "ipf global mutex rwlock");
-	RWLOCK_INIT(&ipf_ipidfrag, "ipf IP NAT-Frag rwlock");
-
-	for (i = 0; i < sizeof(ipf_hooks)/sizeof(ipf_hooks[0]); i++) {
-		err = nf_register_hook(&ipf_hooks[i]);
-		if (err != 0)
-			return err;
-	}
-
-	if (fr_initialise() == -1) {
-		for (i = 0; i < sizeof(ipf_hooks)/sizeof(ipf_hooks[0]); i++)
-			nf_unregister_hook(&ipf_hooks[i]);
-		SPL_X(s);
-		return EIO;
-	}
-
-	bzero((char *)frcache, sizeof(frcache));
-#ifdef notyet
-	if (fr_control_forwarding & 1)
-		ipv4_devconf.forwarding = 1;
-#endif
-
-	SPL_X(s);
-	/* timeout(fr_slowtimer, NULL, (hz / IPF_HZ_DIVIDE) * IPF_HZ_MULT); */
-	return 0;
-}
-
-
-int ipldetach()
-{
-	int i;
-
-	if (fr_refcnt)
-		return EBUSY;
-	SPL_NET(s);
-
-	for (i = 0; i < sizeof(ipf_hooks)/sizeof(ipf_hooks[0]); i++)
-		nf_unregister_hook(&ipf_hooks[i]);
-	/* untimeout(fr_slowtimer, NULL); */
-
-#ifdef notyet
-	if (fr_control_forwarding & 2)
-		ipv4_devconf.forwarding = 0;
-#endif
-
-	fr_deinitialise();
-
-	(void) frflush(IPL_LOGIPF, FR_INQUE|FR_OUTQUE|FR_INACTIVE);
-	(void) frflush(IPL_LOGIPF, FR_INQUE|FR_OUTQUE);
-
-	MUTEX_DESTROY(&ipf_timeoutlock);
-	MUTEX_DESTROY(&ipl_mutex);
-	MUTEX_DESTROY(&ipf_rw);
-	RW_DESTROY(&ipf_mutex);
-	RW_DESTROY(&ipf_global);
-	RW_DESTROY(&ipf_ipidfrag);
-
-	SPL_X(s);
-
-	return 0;
-}
-
-
 static u_int ipf_linux_inout(hooknum, skbp, inifp, outifp, okfn)
 u_int hooknum;
 struct sk_buff **skbp;
@@ -696,6 +926,7 @@ int (*okfn)(struct sk_buff *);
 
 	if (result != 0)
 		return NF_DROP;
+
 	if (ip->ip_v == 4) {
 		ip->ip_len = htons(ip->ip_len);
 		ip->ip_off = htons(ip->ip_off);
@@ -707,7 +938,7 @@ int (*okfn)(struct sk_buff *);
 INLINE void ipf_read_enter(rwlk)
 ipfrwlock_t *rwlk;
 {
-#ifdef IPFDEBUG
+#if defined(IPFDEBUG) && !defined(_KERNEL)
 	if (rwlk->ipf_magic != 0x97dd8b3a) {
 		printk("ipf_read_enter:rwlk %p ipf_magic 0x%x\n",
 			rwlk, rwlk->ipf_magic);
@@ -723,7 +954,7 @@ ipfrwlock_t *rwlk;
 INLINE void ipf_write_enter(rwlk)
 ipfrwlock_t *rwlk;
 {
-#ifdef IPFDEBUG
+#if defined(IPFDEBUG) && !defined(_KERNEL)
 	if (rwlk->ipf_magic != 0x97dd8b3a) {
 		printk("ipf_write_enter:rwlk %p ipf_magic 0x%x\n",
 			rwlk, rwlk->ipf_magic);
@@ -739,7 +970,7 @@ ipfrwlock_t *rwlk;
 INLINE void ipf_rw_exit(rwlk)
 ipfrwlock_t *rwlk;
 {
-#ifdef IPFDEBUG
+#if defined(IPFDEBUG) && !defined(_KERNEL)
 	if (rwlk->ipf_magic != 0x97dd8b3a) {
 		printk("ipf_rw_exit:rwlk %p ipf_magic 0x%x\n",
 			rwlk, rwlk->ipf_magic);
@@ -768,5 +999,145 @@ INLINE void ipf_rw_downgrade(rwlk)
 ipfrwlock_t *rwlk;
 {
 	ipf_rw_exit(rwlk);
-	ipf_write_enter(rwlk);
+	ipf_read_enter(rwlk);
+}
+
+
+#if 0
+void dumpskbuff(sk)
+struct sk_buff *sk;
+{
+	char line[80], *t;
+	u_char *s1, *s2;
+	int len, i;
+	u_char c;
+
+	while (sk != NULL) {
+		len = sk->end - sk->data;
+
+		for (s1 = MTOD(sk, char *); len > 0; ) {
+			t = line;
+			s2 = s1;
+			for (i = 0; i < len && i < 16; i++)  {
+				c = *s2++;
+				sprintf(t, "%02x", c);
+				t += 2;
+				if (i &1)
+					*t++ = ' ';
+			}
+			*t++ = ' ';
+			*t++ = ' ';
+			for (i = 0; i < len && i < 16; i++)  {
+				c = *s1++;
+				*t++ = (c >= 32 && c < 127) ? c : '.';
+			}
+			*t = '\0';
+			printf("%p@%03d %s\n", sk, s1 - MTOD(sk, u_char *), line);
+			if (len > 16) {
+				len -= 16;
+				s1 += 16;
+			} else
+				len = 0;
+		}
+		sk = sk->next;
+	}
+}
+#endif
+
+
+mb_t *m_pullup(m, len)
+mb_t *m;
+int len;
+{
+	if (len <= M_LEN(m))
+		return m;
+	kfree_skb(m);
+	return NULL;
+}
+
+
+/* ------------------------------------------------------------------------ */
+/* Function:    fr_pullup                                                   */
+/* Returns:     NULL == pullup failed, else pointer to protocol header      */
+/* Parameters:  m(I)   - pointer to buffer where data packet starts         */
+/*              fin(I) - pointer to packet information                      */
+/*              len(I) - number of bytes to pullup                          */
+/*                                                                          */
+/* Attempt to move at least len bytes (from the start of the buffer) into a */
+/* single buffer for ease of access.  Operating system native functions are */
+/* used to manage buffers - if necessary.  If the entire packet ends up in  */
+/* a single buffer, set the FI_COALESCE flag even though fr_coalesce() has  */
+/* not been called.  Both fin_ip and fin_dp are updated before exiting _IF_ */
+/* and ONLY if the pullup succeeds.                                         */
+/*                                                                          */
+/* We assume that 'min' is a pointer to a buffer that is part of the chain  */
+/* of buffers that starts at *fin->fin_mp.                                  */
+/* ------------------------------------------------------------------------ */
+void *fr_pullup(min, fin, len)
+mb_t *min;
+fr_info_t *fin;
+int len;
+{
+	int out = fin->fin_out, dpoff, ipoff;
+	mb_t *m = min;
+	char *ip;
+
+	if (m == NULL)
+		return NULL;
+
+	ip = (char *)fin->fin_ip;
+	if ((fin->fin_flx & FI_COALESCE) != 0)
+		return ip;
+
+	ipoff = fin->fin_ipoff;
+	if (fin->fin_dp != NULL)
+		dpoff = (char *)fin->fin_dp - (char *)ip;
+	else
+		dpoff = 0;
+
+	if (M_LEN(m) < len) {
+		m = m_pullup(m, len);
+		*fin->fin_mp = m;
+		fin->fin_m = m;
+		if (m == NULL) {
+			ATOMIC_INCL(frstats[out].fr_pull[1]);
+			return NULL;
+		}
+		ip = MTOD(m, char *) + ipoff;
+	}
+
+	ATOMIC_INCL(frstats[out].fr_pull[0]);
+	fin->fin_ip = (ip_t *)ip;
+	if (fin->fin_dp != NULL)
+		fin->fin_dp = (char *)fin->fin_ip + dpoff;
+
+	if (len == fin->fin_plen)
+		fin->fin_flx |= FI_COALESCE;
+	return ip;
+}
+
+
+/* ------------------------------------------------------------------------ */
+/* Function:    fr_slowtimer                                                */
+/* Returns:     Nil                                                         */
+/* Parameters:  Nil                                                         */
+/*                                                                          */
+/* Slowly expire held state for fragments.  Timeouts are set * in           */
+/* expectation of this being called twice per second.                       */
+/* ------------------------------------------------------------------------ */
+void fr_slowtimer(long value)
+{
+	READ_ENTER(&ipf_global);
+
+	fr_fragexpire();
+	fr_timeoutstate();
+	fr_natexpire();
+	fr_authexpire();
+	fr_ticks++;
+	if (fr_running <= 0)
+		goto done;
+	mod_timer(&ipf_timer, HZ/2 + jiffies);
+
+done:
+	RWLOCK_EXIT(&ipf_global);
 }
