@@ -14,7 +14,9 @@
 #include <sys/param.h>
 #include <sys/time.h>
 #include <sys/file.h>
-#include <sys/fcntl.h>
+#if !defined(AIX)
+# include <sys/fcntl.h>
+#endif
 #if !defined(_KERNEL) && !defined(__KERNEL__)
 # include <stdio.h>
 # include <string.h>
@@ -33,7 +35,8 @@ struct file;
 #include <sys/socket.h>
 #if defined(_KERNEL)
 # if !defined(__NetBSD__) && !defined(sun) && !defined(__osf__) && \
-     !defined(__OpenBSD__) && !defined(__hpux) && !defined(__sgi)
+     !defined(__OpenBSD__) && !defined(__hpux) && !defined(__sgi) && \
+     !defined(AIX)
 #  include <sys/ctype.h>
 # endif
 # include <sys/systm.h>
@@ -86,14 +89,12 @@ struct file;
 #endif
 
 #include "netinet/ip_ftp_pxy.c"
+#include "netinet/ip_rcmd_pxy.c"
+# include "netinet/ip_pptp_pxy.c"
 #if defined(_KERNEL)
 # include "netinet/ip_irc_pxy.c"
-# include "netinet/ip_rcmd_pxy.c"
 # include "netinet/ip_raudio_pxy.c"
 # include "netinet/ip_h323_pxy.c"
-# ifdef	IPFILTER_PRO
-#  include "netinet/ip_msnrpc_pxy.c"
-# endif
 # include "netinet/ip_netbios_pxy.c"
 #endif
 #include "netinet/ip_ipsec_pxy.c"
@@ -102,13 +103,18 @@ struct file;
 /* END OF INCLUDES */
 
 #if !defined(lint)
-static const char rcsid[] = "@(#)$Id: ip_proxy.c,v 2.62 2004/01/08 13:34:29 darrenr Exp $";
+static const char rcsid[] = "@(#)$Id: ip_proxy.c,v 2.62.2.14 2005/06/18 02:41:33 darrenr Exp $";
 #endif
 
 static int appr_fixseqack __P((fr_info_t *, ip_t *, ap_session_t *, int ));
 
 #define	AP_SESS_SIZE	53
 
+#if defined(_KERNEL)
+int		ipf_proxy_debug = 0;
+#else
+int		ipf_proxy_debug = 2;
+#endif
 ap_session_t	*ap_sess_tab[AP_SESS_SIZE];
 ap_session_t	*ap_sess_list = NULL;
 aproxy_t	*ap_proxylist = NULL;
@@ -140,7 +146,12 @@ aproxy_t	ap_proxies[] = {
 #ifdef	IPF_IPSEC_PROXY
 	{ NULL, "ipsec", (char)IPPROTO_UDP, 0, 0,
 	  ippr_ipsec_init, ippr_ipsec_fini, ippr_ipsec_new, ippr_ipsec_del,
-	  NULL, ippr_ipsec_out, NULL, NULL },
+	  ippr_ipsec_inout, ippr_ipsec_inout, ippr_ipsec_match, NULL },
+#endif
+#ifdef	IPF_PPTP_PROXY
+	{ NULL, "pptp", (char)IPPROTO_TCP, 0, 0,
+	  ippr_pptp_init, ippr_pptp_fini, ippr_pptp_new, ippr_pptp_del,
+	  ippr_pptp_inout, ippr_pptp_inout, NULL, NULL },
 #endif
 #ifdef  IPF_H323_PROXY
 	{ NULL, "h323", (char)IPPROTO_TCP, 0, 0, ippr_h323_init, ippr_h323_fini,
@@ -173,14 +184,22 @@ aproxy_t *ap;
 	for (a = ap_proxies; a->apr_p; a++)
 		if ((a->apr_p == ap->apr_p) &&
 		    !strncmp(a->apr_label, ap->apr_label,
-			     sizeof(ap->apr_label)))
+			     sizeof(ap->apr_label))) {
+			if (ipf_proxy_debug > 1)
+				printf("appr_add: %s/%d already present (B)\n",
+				       a->apr_label, a->apr_p);
 			return -1;
+		}
 
 	for (a = ap_proxylist; a->apr_p; a = a->apr_next)
 		if ((a->apr_p == ap->apr_p) &&
 		    !strncmp(a->apr_label, ap->apr_label,
-			     sizeof(ap->apr_label)))
+			     sizeof(ap->apr_label))) {
+			if (ipf_proxy_debug > 1)
+				printf("appr_add: %s/%d already present (D)\n",
+				       a->apr_label, a->apr_p);
 			return -1;
+		}
 	ap->apr_next = ap_proxylist;
 	ap_proxylist = ap;
 	if (ap->apr_init != NULL)
@@ -201,12 +220,22 @@ ap_ctl_t *ctl;
 	int error;
 
 	a = appr_lookup(ctl->apc_p, ctl->apc_label);
-	if (a == NULL)
+	if (a == NULL) {
+		if (ipf_proxy_debug > 1)
+			printf("appr_ctl: can't find %s/%d\n",
+				ctl->apc_label, ctl->apc_p);
 		error = ESRCH;
-	else if (a->apr_ctl == NULL)
+	} else if (a->apr_ctl == NULL) {
+		if (ipf_proxy_debug > 1)
+			printf("appr_ctl: no ctl function for %s/%d\n",
+				ctl->apc_label, ctl->apc_p);
 		error = ENXIO;
-	else
+	} else {
 		error = (*a->apr_ctl)(a, ctl);
+		if ((error != 0) && (ipf_proxy_debug > 1))
+			printf("appr_ctl: %s/%d ctl error %d\n",
+				a->apr_label, a->apr_p, error);
+	}
 	return error;
 }
 
@@ -225,10 +254,16 @@ aproxy_t *ap;
 		if (a == ap) {
 			a->apr_flags |= APR_DELETE;
 			*app = a->apr_next;
-			if (ap->apr_ref != 0)
+			if (ap->apr_ref != 0) {
+				if (ipf_proxy_debug > 2)
+					printf("appr_del: orphaning %s/%d\n",
+						ap->apr_label, ap->apr_p);
 				return 1;
+			}
 			return 0;
 		}
+	if (ipf_proxy_debug > 1)
+		printf("appr_del: proxy %lx not found\n", (u_long)ap);
 	return -1;
 }
 
@@ -288,7 +323,8 @@ int mode;
 		if (error == 0)
 			error = appr_ctl(&ctl);
 
-		if (ctl.apc_dsize > 0 && ptr != NULL && ctl.apc_data == ptr) {
+		if ((ctl.apc_dsize > 0) && (ptr != NULL) &&
+		    (ctl.apc_data == ptr)) {
 			KFREES(ptr, ctl.apc_dsize);
 		}
 		break;
@@ -310,17 +346,37 @@ nat_t *nat;
 {
 	aproxy_t *apr;
 	ipnat_t *ipn;
+	int result;
 
 	ipn = nat->nat_ptr;
-	if (ipn == NULL)
+	if (ipf_proxy_debug > 8)
+		printf("appr_match(%lx,%lx) aps %lx ptr %lx\n",
+			(u_long)fin, (u_long)nat, (u_long)nat->nat_aps,
+			(u_long)ipn);
+
+	if ((fin->fin_flx & (FI_SHORT|FI_BAD)) != 0) {
+		if (ipf_proxy_debug > 0)
+			printf("appr_match: flx 0x%x (BAD|SHORT)\n",
+				fin->fin_flx);
 		return -1;
+	}
+
 	apr = ipn->in_apr;
-	if ((apr == NULL) || (apr->apr_flags & APR_DELETE) ||
-	    (nat->nat_aps == NULL))
+	if ((apr == NULL) || (apr->apr_flags & APR_DELETE)) {
+		if (ipf_proxy_debug > 0)
+			printf("appr_match:apr %lx apr_flags 0x%x\n",
+				(u_long)apr, apr ? apr->apr_flags : 0);
 		return -1;
-	if (apr->apr_match != NULL)
-		if ((*apr->apr_match)(fin, nat->nat_aps, nat) != 0)
+	}
+
+	if (apr->apr_match != NULL) {
+		result = (*apr->apr_match)(fin, nat->nat_aps, nat);
+		if (result != 0) {
+			if (ipf_proxy_debug > 4)
+				printf("appr_match: result %d\n", result);
 			return -1;
+		}
+	}
 	return 0;
 }
 
@@ -337,18 +393,34 @@ nat_t *nat;
 	register ap_session_t *aps;
 	aproxy_t *apr;
 
-	if ((nat->nat_ptr == NULL) || (nat->nat_aps != NULL))
+	if (ipf_proxy_debug > 8)
+		printf("appr_new(%lx,%lx) \n", (u_long)fin, (u_long)nat);
+
+	if ((nat->nat_ptr == NULL) || (nat->nat_aps != NULL)) {
+		if (ipf_proxy_debug > 0)
+			printf("appr_new: nat_ptr %lx nat_aps %lx\n",
+				(u_long)nat->nat_ptr, (u_long)nat->nat_aps);
 		return -1;
+	}
 
 	apr = nat->nat_ptr->in_apr;
 
-	if (!apr || (apr->apr_flags & APR_DELETE) ||
-	    (fin->fin_p != apr->apr_p))
+	if ((apr->apr_flags & APR_DELETE) ||
+	    (fin->fin_p != apr->apr_p)) {
+		if (ipf_proxy_debug > 2)
+			printf("appr_new: apr_flags 0x%x p %d/%d\n",
+				apr->apr_flags, fin->fin_p, apr->apr_p);
 		return -1;
+	}
 
 	KMALLOC(aps, ap_session_t *);
-	if (!aps)
+	if (!aps) {
+		if (ipf_proxy_debug > 0)
+			printf("appr_new: malloc failed (%lu)\n",
+				(u_long)sizeof(ap_session_t));
 		return -1;
+	}
+
 	bzero((char *)aps, sizeof(*aps));
 	aps->aps_p = fin->fin_p;
 	aps->aps_data = NULL;
@@ -360,6 +432,9 @@ nat_t *nat;
 				KFREES(aps->aps_data, aps->aps_psiz);
 			}
 			KFREE(aps);
+			if (ipf_proxy_debug > 2)
+				printf("appr_new: new(%lx) failed\n",
+					(u_long)apr->apr_new);
 			return -1;
 		}
 	aps->aps_nat = nat;
@@ -398,14 +473,17 @@ nat_t *nat;
 	u_32_t s1, s2, sd;
 #endif
 
-	if (fin->fin_flx & FI_BAD)
+	if (fin->fin_flx & FI_BAD) {
+		if (ipf_proxy_debug > 0)
+			printf("appr_check: flx 0x%x (BAD)\n", fin->fin_flx);
 		return -1;
+	}
 
 #ifndef IPFILTER_CKSUM
 	if ((fin->fin_out == 0) && (fr_checkl4sum(fin) == -1)) {
-# if PROXY_DEBUG || !defined(_KERNEL)
-		printf("proxy l4 checksum failure\n");
-# endif
+		if (ipf_proxy_debug > 0)
+			printf("appr_check: l4 checksum failure %d\n",
+				fin->fin_p);
 		if (fin->fin_p == IPPROTO_TCP)
 			frstats[fin->fin_out].fr_tcpbad++;
 		return -1;
@@ -418,9 +496,14 @@ nat_t *nat;
 		 * If there is data in this packet to be proxied then try and
 		 * get it all into the one buffer, else drop it.
 		 */
+#if defined(MENTAT) || defined(HAVE_M_PULLDOWN)
 		if ((fin->fin_dlen > 0) && !(fin->fin_flx & FI_COALESCE))
-			if (fr_coalesce(fin) == -1)
+			if (fr_coalesce(fin) == -1) {
+				if (ipf_proxy_debug > 0)
+					printf("appr_check: fr_coalesce failed %x\n", fin->fin_flx);
 				return -1;
+			}
+#endif
 		ip = fin->fin_ip;
 
 		switch (fin->fin_p)
@@ -458,16 +541,14 @@ nat_t *nat;
 		}
 
 		rv = APR_EXIT(err);
-		if (rv == 1) {
-#if PROXY_DEBUG || !defined(_KERNEL)
-			printf("proxy says bad packet received\n");
-#endif
+		if (((ipf_proxy_debug > 0) && (rv != 0)) ||
+		    (ipf_proxy_debug > 8))
+			printf("appr_check: out %d err %x rv %d\n",
+				fin->fin_out, err, rv);
+		if (rv == 1)
 			return -1;
-		}
+
 		if (rv == 2) {
-#if PROXY_DEBUG || !defined(_KERNEL)
-			printf("proxy says free app proxy data\n");
-#endif
 			appr_free(apr);
 			nat->nat_aps = NULL;
 			return -1;
@@ -535,6 +616,9 @@ char *name;
 {
 	aproxy_t *ap;
 
+	if (ipf_proxy_debug > 8)
+		printf("appr_lookup(%d,%s)\n", pr, name);
+
 	for (ap = ap_proxies; ap->apr_p; ap++)
 		if ((ap->apr_p == pr) &&
 		    !strncmp(name, ap->apr_label, sizeof(ap->apr_label))) {
@@ -548,6 +632,8 @@ char *name;
 			ap->apr_ref++;
 			return ap;
 		}
+	if (ipf_proxy_debug > 2)
+		printf("appr_lookup: failed for %d/%s\n", pr, name);
 	return NULL;
 }
 
@@ -616,10 +702,10 @@ int inc;
 		/* switch to other set ? */
 		if ((aps->aps_seqmin[!sel] > aps->aps_seqmin[sel]) &&
 		    (seq1 > aps->aps_seqmin[!sel])) {
-#if PROXY_DEBUG
-			printf("proxy out switch set seq %d -> %d %x > %x\n",
-				sel, !sel, seq1, aps->aps_seqmin[!sel]);
-#endif
+			if (ipf_proxy_debug > 7)
+				printf("proxy out switch set seq %d -> %d %x > %x\n",
+					sel, !sel, seq1,
+					aps->aps_seqmin[!sel]);
 			sel = aps->aps_sel[out] = !sel;
 		}
 
@@ -636,11 +722,10 @@ int inc;
 		if (inc && (seq1 > aps->aps_seqmin[!sel])) {
 			aps->aps_seqmin[sel] = seq1 + nlen - 1;
 			aps->aps_seqoff[sel] = aps->aps_seqoff[sel] + inc;
-#if PROXY_DEBUG
-			printf("proxy seq set %d at %x to %d + %d\n", sel,
-				aps->aps_seqmin[sel], aps->aps_seqoff[sel],
-				inc);
-#endif
+			if (ipf_proxy_debug > 7)
+				printf("proxy seq set %d at %x to %d + %d\n",
+					sel, aps->aps_seqmin[sel],
+					aps->aps_seqoff[sel], inc);
 		}
 
 		/***/
@@ -651,10 +736,10 @@ int inc;
 		/* switch to other set ? */
 		if ((aps->aps_ackmin[!sel] > aps->aps_ackmin[sel]) &&
 		    (seq1 > aps->aps_ackmin[!sel])) {
-#if PROXY_DEBUG
-			printf("proxy out switch set ack %d -> %d %x > %x\n",
-				sel, !sel, seq1, aps->aps_ackmin[!sel]);
-#endif
+			if (ipf_proxy_debug > 7)
+				printf("proxy out switch set ack %d -> %d %x > %x\n",
+					sel, !sel, seq1,
+					aps->aps_ackmin[!sel]);
 			sel = aps->aps_sel[1 - out] = !sel;
 		}
 
@@ -670,10 +755,9 @@ int inc;
 		/* switch to other set ? */
 		if ((aps->aps_ackmin[!sel] > aps->aps_ackmin[sel]) &&
 		    (seq1 > aps->aps_ackmin[!sel])) {
-#if PROXY_DEBUG
-			printf("proxy in switch set ack %d -> %d %x > %x\n",
-				sel, !sel, seq1, aps->aps_ackmin[!sel]);
-#endif
+			if (ipf_proxy_debug > 7)
+				printf("proxy in switch set ack %d -> %d %x > %x\n",
+					sel, !sel, seq1, aps->aps_ackmin[!sel]);
 			sel = aps->aps_sel[out] = !sel;
 		}
 
@@ -690,11 +774,11 @@ int inc;
 		if (inc && (seq1 > aps->aps_ackmin[!sel])) {
 			aps->aps_ackmin[!sel] = seq1 + nlen - 1;
 			aps->aps_ackoff[!sel] = aps->aps_ackoff[sel] + inc;
-#if PROXY_DEBUG
-			printf("proxy ack set %d at %x to %d + %d\n", !sel,
-				aps->aps_seqmin[!sel], aps->aps_seqoff[sel],
-				inc);
-#endif
+
+			if (ipf_proxy_debug > 7)
+				printf("proxy ack set %d at %x to %d + %d\n",
+					!sel, aps->aps_seqmin[!sel],
+					aps->aps_seqoff[sel], inc);
 		}
 
 		/***/
@@ -705,19 +789,17 @@ int inc;
 		/* switch to other set ? */
 		if ((aps->aps_seqmin[!sel] > aps->aps_seqmin[sel]) &&
 		    (seq1 > aps->aps_seqmin[!sel])) {
-#if PROXY_DEBUG
-			printf("proxy in switch set seq %d -> %d %x > %x\n",
-				sel, !sel, seq1, aps->aps_seqmin[!sel]);
-#endif
+			if (ipf_proxy_debug > 7)
+				printf("proxy in switch set seq %d -> %d %x > %x\n",
+					sel, !sel, seq1, aps->aps_seqmin[!sel]);
 			sel = aps->aps_sel[1 - out] = !sel;
 		}
 
 		if (aps->aps_seqoff[sel] != 0) {
-#if PROXY_DEBUG
-			printf("sel %d seqoff %d seq1 %x seqmin %x\n", sel,
-				aps->aps_seqoff[sel], seq1,
-				aps->aps_seqmin[sel]);
-#endif
+			if (ipf_proxy_debug > 7)
+				printf("sel %d seqoff %d seq1 %x seqmin %x\n",
+					sel, aps->aps_seqoff[sel], seq1,
+					aps->aps_seqmin[sel]);
 			if (seq1 > aps->aps_seqmin[sel]) {
 				seq2 = aps->aps_seqoff[sel];
 				tcp->th_ack = htonl(seq1 - seq2);
@@ -725,10 +807,10 @@ int inc;
 			}
 		}
 	}
-#if PROXY_DEBUG
-	printf("appr_fixseqack: seq %lx ack %lx\n", ntohl(tcp->th_seq),
-		ntohl(tcp->th_ack));
-#endif
+
+	if (ipf_proxy_debug > 8)
+		printf("appr_fixseqack: seq %x ack %x\n",
+			ntohl(tcp->th_seq), ntohl(tcp->th_ack));
 	return ch ? 2 : 0;
 }
 
