@@ -6,7 +6,7 @@
  */
 #if !defined(lint)
 static const char sccsid[] = "@(#)sock.c	1.2 1/11/96 (C)1995 Darren Reed";
-static const char rcsid[] = "@(#)$Id: sock.c,v 2.8 2002/02/24 07:30:03 darrenr Exp $";
+static const char rcsid[] = "@(#)$Id: sock.c,v 2.8.4.1 2004/03/23 12:58:06 darrenr Exp $";
 #endif
 #include <sys/param.h>
 #include <sys/types.h>
@@ -54,6 +54,9 @@ static const char rcsid[] = "@(#)$Id: sock.c,v 2.8 2002/02/24 07:30:03 darrenr E
 #include <netinet/ip.h>
 #include <netinet/tcp.h>
 #include <net/if.h>
+#if defined(__FreeBSD__)
+# include "radix_ipf.h"
+#endif
 #include <net/route.h>
 #include <netinet/ip_var.h>
 #include <netinet/in_pcb.h>
