@@ -3,7 +3,7 @@
  *
  * See the IPFILTER.LICENCE file for details on licencing.
  *
- * $Id: icmpcode.c,v 1.7 2003/12/01 01:59:07 darrenr Exp $
+ * $Id: icmpcode.c,v 1.7.2.1 2004/12/09 19:41:20 darrenr Exp $
  */
 
 #include <ctype.h>
@@ -32,7 +32,7 @@ char *str;
 
 	if ((s = strrchr(str, ')')))
 		*s = '\0';
-	if (isdigit(*str)) {
+	if (ISDIGIT(*str)) {
 		if (!ratoi(str, &i, 0, 255))
 			return -1;
 		else
