@@ -40,7 +40,7 @@
 
 #if !defined(lint)
 static const char sccsid[] = "@(#)mls_ipl.c	2.6 10/15/95 (C) 1993-2000 Darren Reed";
-static const char rcsid[] = "@(#)$Id: mls_ipl.c,v 2.7 2004/02/07 17:27:33 darrenr Exp $";
+static const char rcsid[] = "@(#)$Id: mls_ipl.c,v 2.7.2.2 2004/11/04 19:24:54 darrenr Exp $";
 #endif
 
 extern	int	ipldetach __P((void));
@@ -76,7 +76,7 @@ struct	dev_ops	ipl_ops =
 	iplopen,
 	iplclose,
 	iplread,
-	NULL,		/* write */
+	iplwrite,
 	NULL,		/* strategy */
 	NULL,		/* dump */
 	0,		/* psize */
@@ -171,12 +171,16 @@ struct	vdstat	*vds;
 
 static	int	unload()
 {
+	int err = 0, i;
 	char *name;
-	int err, i;
 
-	err = ipldetach();
+	if (fr_refcnt != 0)
+		err = EBUSY;
+	else if (fr_running >= 0)
+		err = ipldetach();
 	if (err)
 		return err;
+
 	fr_running = -2;
 	for (i = 0; (name = ipf_devfiles[i]); i++)
 		(void) vn_remove(name, UIO_SYSSPACE, FILE);
