@@ -4,7 +4,7 @@
  * See the IPFILTER.LICENCE file for details on licencing.
  */
 #if !defined(lint)
-static const char rcsid[] = "@(#)$Id: tru64.c,v 2.11 2004/01/04 09:19:17 darrenr Exp $";
+static const char rcsid[] = "@(#)$Id: tru64.c,v 2.11.2.5 2005/07/13 21:40:48 darrenr Exp $";
 #endif
 #include <sys/types.h>
 #include <sys/conf.h>
@@ -175,14 +175,15 @@ struct firewall_stat {
 struct firewall_stat ipfilter_stat;
 static	int	ipfilter_registered = 0;
 static	int	ipftru64_inited = 0;
+static	char	ipfname[CFG_ATTR_NAME_SZ] = "ipfilter";
 
 
 cfg_subsys_attr_t ipfilter_attributes[] = {
 
 { "Subsystem_Description",CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
 				(caddr_t)ipfilter_desc, 2, 300, 0 },
-{ "Module_Config_Name",	CFG_ATTR_STRTYPE, CFG_OP_QUERY,
-				(caddr_t)"ipfilter", 2, CFG_ATTR_NAME_SZ, 0 },
+{ "Module_Config_Name",	CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE,
+				(caddr_t)ipfname, 2, CFG_ATTR_NAME_SZ, 0 },
 { "majnum",		CFG_ATTR_INTTYPE, CFG_OP_QUERY,
 				(caddr_t)&majnum, 0, 512, 0 },
 { "Num_Installed",	CFG_ATTR_INTTYPE, CFG_OP_QUERY,
@@ -193,6 +194,15 @@ cfg_subsys_attr_t ipfilter_attributes[] = {
 { "Module_Type",	CFG_ATTR_STRTYPE, CFG_OP_QUERY | CFG_OP_CONFIGURE,
 				(caddr_t)ipfilter_unused,
 				2, CFG_ATTR_NAME_SZ, 0 },
+{ "Device_User",	CFG_ATTR_STRTYPE, CFG_OP_QUERY | CFG_OP_CONFIGURE,
+				(caddr_t)ipfilter_unused,
+				0, sizeof(ipfilter_unused), 0 },
+{ "Device_Group",	CFG_ATTR_STRTYPE, CFG_OP_QUERY | CFG_OP_CONFIGURE,
+				(caddr_t)ipfilter_unused,
+				0, sizeof(ipfilter_unused), 0 },
+{ "Device_Mode",	CFG_ATTR_STRTYPE, CFG_OP_QUERY | CFG_OP_CONFIGURE,
+				(caddr_t)ipfilter_unused,
+				0, sizeof(ipfilter_unused), 0 },
 { "Device_Major_Req",	CFG_ATTR_STRTYPE, CFG_OP_QUERY | CFG_OP_CONFIGURE,
 				(caddr_t)ipfilter_unused,
 				0, sizeof(ipfilter_unused), 0 },
@@ -223,9 +233,6 @@ cfg_subsys_attr_t ipfilter_attributes[] = {
 				(caddr_t)&fr_minttl, 0, 1, 0 },
 { "fr_running",		CFG_ATTR_INTTYPE, CFG_OP_QUERY,
 				(caddr_t)&fr_minttl, 0, 1, 0 },
-{ "fr_unreach",		CFG_ATTR_INTTYPE, CFG_OP_QUERY |
-				CFG_OP_CONFIGURE | CFG_OP_RECONFIGURE,
-				(caddr_t) &fr_unreach, 0, 255, 0 },
 { "fr_control_forwarding",
 			CFG_ATTR_INTTYPE, CFG_OP_QUERY |
 				CFG_OP_CONFIGURE | CFG_OP_RECONFIGURE,
@@ -271,6 +278,10 @@ ulong outdata_size;
 	switch (op)
 	{
 	case CFG_OP_CONFIGURE :
+#ifdef IPFDEBUG
+		printf("ipfilter_config=%d\n",ipfilter_config);
+		printf("driver_cfg_state=%d\n",driver_cfg_state);
+#endif
 		if (ipfilter_config == TRUE) {
 			/*
 			 * We have already been configured
@@ -342,6 +353,9 @@ ulong outdata_size;
 				return (status);
 			}
 			status = ipfilter_attach();
+#ifdef IPFDEBUG
+			printf("ipfilter_attach=%d\n",status);
+#endif
 		}
 		break;
 
@@ -381,6 +395,9 @@ ulong outdata_size;
 		break;
 	}
 
+#ifdef IPFDEBUG
+	printf("ipfilter_configure(%d)=%d\n",op,status);
+#endif
 	return status;
 }
 
@@ -589,7 +606,7 @@ struct ifnet *ifp;
 	}
 
 	if (cmd == (ioctlcmd_t)SIOCAIFADDR)
-		frsync();
+		frsync(NULL);
 
 	RWLOCK_EXIT(&ipf_tru64);
 }
@@ -618,14 +635,17 @@ ipfilter_attach(void)
 	RWLOCK_INIT(&ipf_tru64, 1);
 	RWLOCK_INIT(&ipf_global, 1);
 	RWLOCK_INIT(&ipf_mutex, 1);
+	RWLOCK_INIT(&ipf_frcache, 1);
 	ipftru64_inited = 1;
 
 	status = iplattach();
 #ifdef	IPFDEBUG
 	printf("iplattach() = %d\n", status);
 #endif
-	if (status != ESUCCESS)
+	if (status != ESUCCESS) {
+		(void) ipldetach();
 		return status;
+	}
 
 	ipfilter_registered = 1;
 
@@ -781,6 +801,7 @@ ipfilter_detach(void)
 		if (ipftru64_inited == 1) {
 			RW_DESTROY(&ipf_tru64);
 			RW_DESTROY(&ipf_global);
+			RW_DESTROY(&ipf_frcache);
 			RW_DESTROY(&ipf_mutex);
 			ipftru64_inited = 0;
 		}
@@ -936,6 +957,7 @@ ipfilter_preconfig_callback(int point, int order, ulong argument, ulong event_ar
 	int status;
 	struct hwconfig *hwc;
 	hwc = create_hwconfig_struct();
+
 	if (hwc == NULL)
 		return;
 
@@ -1116,6 +1138,10 @@ int ipfilterioctl(dev_t dev, unsigned int cmd, caddr_t data, int flag)
 	READ_ENTER(&ipf_tru64);
 	if (ipfilter_registered < 1 || ipftru64_inited == 0) {
 		RWLOCK_EXIT(&ipf_tru64);
+#ifdef IPFDEBUG
+		printf("ipfilter_registered %d ipftru64_inited %d\n",
+			ipfilter_registered, ipftru64_inited);
+#endif
 		return EIO;
 	}
 	err = iplioctl(dev, cmd, data, flag);
